/** \file Connections are segments joined by 2 points. The node array is
 * twice as large as the number of connections. For connection i, the nodes
 * defining it are located in _node[2*i] and node[2+i=1]
 */
#include "DrawingObjects.h"

#include "VecData.h"
#ifdef USE_HDF5
#include <ch5/ch5.h>
#endif


void
Connection::buffer( RenderLines& rd, const GLfloat *col, Colourscale* cs,
        DATA_TYPE *data, dataOpac* dop )
{
  int        ln      = 0;
  GLfloat    nrml[]  = { 0, 0, 1 };
  GLfloat   *cpubuff = NULL;
  bool       translucent = false;

  for( int i=0; i<num(); i++ ) {
    int stop = (ln<_lines.size() && i==_lines[ln].first) ? _lines[ln].second : i;
    int nseg = stop - i + 1;
    GLfloat *buff = cpubuff = (GLfloat *)realloc( cpubuff, (nseg+1)*VBO_ELEM_SZ*sizeof(GLfloat) );
    const GLfloat *c = colour(_node[i*2], col, data, cs, dop );
    buff += buff_vtx( buff, (*_pt)[_node[i*2]], nrml, c, 0 );
    for( ; i<=stop; i++ ){
      c     = colour(_node[i*2+1], col, data, cs, dop );
      if( c[3]<opaque_min ) translucent = true;
      buff += buff_vtx( buff, (*_pt)[_node[i*2+1]], nrml, c, 0 );
    }
    if( nseg>1 ) ln++;
    i--;
    rd.buffer_line( cpubuff, nseg );
  } 
  free( cpubuff );
  rd.opaque(translucent ? 0 : -1);
    
  rd.update_nodalbuff();
}


/** add a list of connections
 *  \param n  number of connections
 *  \param nl node list [c0_0, c0_1, c1_0, c1_1, c2_0, etc]
 */
void Connection:: add( int n, int *nl )
{
  _n += n;
  _node=(int*)realloc(_node, _n*_ptsPerObj*sizeof(int));
  memcpy( _node+(_n-n)*_ptsPerObj, nl, n*sizeof(int)*_ptsPerObj );
}


/** read in the connection file */
bool Connection::read( const char *fname )
{
  gzFile in;

  try {
    in = openFile( fname, "cnnx" );
  } catch (...) { return false; }

  const int bufsize=1024;
  char      buff[bufsize];
  if ( gzgets(in, buff, bufsize) == Z_NULL ) return false;
  if ( sscanf( buff, "%d", &_n ) != 1 ) return false;
  _node = (int*)malloc(2*_n*sizeof(int));
  int start_ln = -1;
  for ( int i=0; i<2*_n; i+=2 ) {
    gzgets(in, buff, bufsize);
    sscanf( buff, "%d %d", _node+i, _node+i+1 );
    if( start_ln>=0 && i && _node[i]!=_node[i-1] ){
      _lines.push_back( make_pair(start_ln, i/2-1) );
      start_ln = -1;
    } else if( start_ln<0 && i && _node[i]==_node[i-1] )
      start_ln =  i/2-1;
  }
  gzclose(in);
  return true;
}

/** read in the connection file */
bool Connection::read_cables( const char *fname )
{
  gzFile in;

  try {
    in = openFile( fname, "cables" );
  } catch (...) { return false; }

  const int bufsize=1024;
  char      buff[bufsize];
  if ( gzgets(in, buff, bufsize) == Z_NULL ) return false;
  int num;
  if ( sscanf( buff, "%d", &num ) != 1 ) return false;

  vector<int>node;
  gzgets(in, buff, bufsize);
  int start, end;
  int c0 = 0;
  sscanf( buff, "%d", &start );
  for ( int i=0; i<num; i++ ) {
    gzgets(in, buff, bufsize);
    sscanf( buff, "%d", &end );
    for( int j=start; j<end-1; j++ ) {
      node.push_back(j);
      node.push_back(j+1);
    }
    _lines.push_back(make_pair(_n+c0,_n+node.size()/2));
    start = end;
    c0 = node.size()/2+1;
  }
  gzclose(in);
  add( node.size()/2, node.data() );
  return true;
}


#ifdef USE_HDF5
bool Connection::read(hid_t hdf_file)
{
  ch5_dataset dset_info;
  if( ch5m_conn_get_info(hdf_file, &dset_info) )
    return false;
  
  _n = dset_info.count;
  _node = (int*)malloc(_n * dset_info.width*sizeof(int));
  if ( ch5m_conn_get_all(hdf_file, _node) ){
    cerr << "Error reading in connections" << endl;
    _n = 0;
    delete _node;
    return false;
  }
  
  return true;
}
#endif
