#ifdef OSMESA
#  include <GL/osmesa.h>
#else
#  ifdef __APPLE__
#    define GL_SILENCE_DEPRECATION
#    include <OpenGL/gl3.h> 
#  else
#    if defined(WIN32)
#      define GLEW_STATIC 1
#    endif
#   include <GL/glew.h>
static bool initglew = false;
#  endif
#endif

#include "Frame.h"

/** constructor to get OpenGL comntext */
Frame :: Frame( TBmeshWin *t ) : _tbwm(t)
{
#if !defined(OSMESA) && !defined(__APPLE__)
  if( !initglew ) {
    initglew = true;
    GLenum err = glewInit();
    if (GLEW_OK != err) {
      cerr << "GLEW problem -- aborting!" << endl;
    }
  }
#endif
}

/** constructor to fill buffer with current screen contents */
Frame :: Frame( TBmeshWin *t, int w, int h, bool fa ) : _tbwm(t), _full_alpha(fa)
{
#if !defined(OSMESA) && !defined(__APPLE__)
  if( !initglew ) {
    initglew = true;
    GLenum err = glewInit();
    if (GLEW_OK != err) {
      cerr << "GLEW problem -- aborting!" << endl;
    }
  }
#endif
  fill_buffer( w, h, true );

#ifdef OSMESA
#else
  glBindFramebuffer(GL_FRAMEBUFFER, 0); // restore the display
#endif //OSMESA
}


/** constructor to dump to a file */
Frame :: Frame( TBmeshWin *t, int w, int h, string fname ) : _tbwm(t)
{
  dump( w, h, fname );
}


Frame :: ~Frame()  
{
  delete_objs();
}


void 
Frame :: delete_objs()
{
#ifdef OSMESA
  OSMesaDestroyContext( _ctx );
#else
  delete[] _buffer;
  glDeleteRenderbuffers(1, &_color_rb);
  glDeleteRenderbuffers(1, &_depth_rb);
  glDeleteFramebuffers(1, &_fb);
#endif // OSMESA
}


/** output specified frames
 *
 * \param w     output image width
 * \param h     output image height
 * \param fname base file name
 * \param f0    first frame number
 * \param f1    largest frame = f0+n-1
 * \param stride output frame number stride
 *
 * \note \p fname will have ".png" removed if present and XXXXX.png 
 *       where XXXXX is the frame number
 *
 * \note the frame numbers written are those which satisfy f0<=n<=f1
 *
 * \return the number of frames written
 */
int
Frame :: write( int w, int h, string fname, int f0, int f1, int stride )
{
  if( !_tbwm->set_time(f0) ) return 0;

  if( fname.length()>=4 && (fname.substr(fname.length()-4)==".png") )
    fname = fname.substr(0,fname.length()-4);

  int numout=0;
  for( int f=f0; f<=f1; f+=stride) {
    if( !_tbwm->set_time(f) )
      break;
    string file=fname;
    char strnum[256];
    sprintf( strnum, "%05d.png", f );
    file += strnum;
    dump( w, h, file );
    numout++;
  }
  return  numout;
}


/** output specified frame
 *
 * \param w     output image width
 * \param h     output image height
 * \param fname file name
 * \param f     frame number
 *
 * \note \p fname will have ".png" appended if not present
 */
int
Frame :: write( int w, int h, string fname, int f )
{
  if( !_tbwm->set_time(f) ) return 0;

  if( fname.length()<4 || (fname.substr(fname.length()-4)!=".png") )
    fname+= ".png";

  dump( w, h, fname );
  return 1;
}


/* dump the frame buffer into a file 
 *
 * \param w
 * \param h
 * \param fname
 */
void 
Frame :: dump( int w, int h, string fname )
{
  FILE *out = fopen( fname.c_str(), "w" );
  PNGwrite* pngimg = new PNGwrite( out );
  pngimg->size( w, h );
  pngimg->depth( 8*sizeof(GLubyte) );

  char *datafile, cscale[1024]={0}, time[32]={0};
  if( _tbwm->dataBuffer ) {
    datafile = realpath( _tbwm->dataBuffer->file().c_str(), NULL );
    sprintf( cscale, "data range = [%f, %f], %d levels", 
            _tbwm->cs->min(), _tbwm->cs->max(), _tbwm->cs->size() );
    if( _tbwm->data ) 
      sprintf( time, "Frame %d --> %f ms",  _tbwm->time(),
               _tbwm->time()*_tbwm->dataBuffer->dt()+_tbwm->dataBuffer->t0() );
    else
      sprintf( time, "Frame 0" );
  } else
    datafile = (char *)calloc(1,1);

  string agrid = (_tbwm->auxGrid!=NULL) ? _tbwm->auxGrid->file() : (char *)calloc(1,1); 
  string vdat  = (_tbwm->vecdata!=NULL) ? _tbwm->vecdata->file() : (char *)calloc(1,1); 
  char mpath[2048];
  realpath( (_tbwm->model->file()+"pts").c_str(), mpath );
  pngimg->description( mpath, datafile, cscale, time, agrid.c_str(), vdat.c_str() );

  free(datafile);

  if( _tbwm->transBgd() )
    pngimg->colour_type( PNG_COLOR_TYPE_RGB_ALPHA );
  else
    pngimg->colour_type( PNG_COLOR_TYPE_RGB );

  fill_buffer( w, h );

  int align;
  glGetIntegerv(GL_PACK_ALIGNMENT, &align);
  pngimg->write( _buffer, align );
  delete pngimg;

#ifdef OSMESA
#else
  glBindFramebuffer(GL_FRAMEBUFFER, 0); // restore the display
#endif //OSMESA
}


void 
Frame::fill_buffer( int w, int h, bool alpha )
{
  if( _buffer==NULL || _w!=w || _h!=h ) {   // resize 
    
    if( _buffer )   // not first time, destroy old buffers/contexts
      delete_objs();
	
    _w = w;
    _h = h;
    _buffer = new GLubyte[w*h*4];

#ifdef OSMESA
    const int attribs[] = { OSMESA_FORMAT,      OSMESA_RGBA, 
                            OSMESA_DEPTH_BITS,           32,
                            OSMESA_STENCIL_BITS,          0,
                            OSMESA_ACCUM_BITS,            0,
                            OSMESA_PROFILE, OSMESA_CORE_PROFILE,
                            OSMESA_CONTEXT_MAJOR_VERSION, 3,
                            OSMESA_CONTEXT_MINOR_VERSION, 3,
                            0 };
    _ctx = OSMesaCreateContextAttribs( attribs, NULL );
    if( !_ctx ) {
      cout << "Failed OSMesaCreateContextAttribs" << endl;
      throw 1;
    }

    if (!OSMesaMakeCurrent(_ctx, _buffer, GL_UNSIGNED_BYTE, w, h)) {
      cout << "Failed OSMesaMakeCurrent" << endl;
      throw 1;
    }
#else 
    // create new frame buffer object

    glGenFramebuffers(1, &_fb);

    glBindFramebuffer(GL_FRAMEBUFFER, _fb);
    glGenRenderbuffers(1, &_color_rb);
    glBindRenderbuffer(GL_RENDERBUFFER, _color_rb);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA8, w, h );
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
            GL_RENDERBUFFER, _color_rb);

    glGenRenderbuffers(1, &_depth_rb);
    glBindRenderbuffer(GL_RENDERBUFFER, _depth_rb);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, w, h );
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
            GL_RENDERBUFFER, _depth_rb);

    GLenum gle=glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if( gle!=GL_FRAMEBUFFER_COMPLETE ) {
      cout << "Frame buffer incomplete -> aborting:" << gle << endl;
      throw 1;
    }
#endif //OSMESA
  }

#ifndef OSMESA
  glBindFramebuffer(GL_FRAMEBUFFER, _fb);
#endif //OSMESA

  _tbwm->invalidate();
  _tbwm->draw();

  glReadBuffer(GL_BACK);
  if( _tbwm->transBgd() || alpha ){
    glReadPixels(0,0,w,h,GL_RGBA,GL_UNSIGNED_BYTE,(GLvoid *)_buffer);
    if( !_full_alpha )
      for( int i=0; i<w*h; i++ )
        if( _buffer[i*4+3] > 0 ) _buffer[i*4+3] = 255;
  } else
    glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid *)_buffer);
}
 
