#include "DrawingObjects.h"
#include "VecData.h"
#include "gzFileBuffer.h"
#include "AsciiReader.hpp"
#include <DataOnDemand.h>
#ifdef USE_HDF5
#include <ch5/ch5.h>
#endif

/** buffer many PPoints
 *
 *  \param rd       render VBO object
 *  \param p0       first index of point to draw
 *  \param p1       last index of point to draw
 *  \param colour   colour to use if no data
 *  \param data     data associated with nodes (NULL for no data display)
 *  \param stride   draw every n'th point
 */
void 
PPoint::buffer( RenderSpheres &rd, int p0, int p1, const GLfloat *defcol, Colourscale* cs,
                  DATA_TYPE* data, int stride, dataOpac* dataopac )
{
  if ( p0>=_n || p1>=_n ) return;

  int nvis=0;
  for( int i=p0; i<=p1; i++ ) if( (*_visible)[i] ) nvis++;
  GLfloat *buff = rd.buff_append( nvis );

  int current=0;
#pragma omp parallel for
  for ( int i=p0; i<=p1; i+=stride ) 
    if ( (*_visible)[i] ) {
      int buffent;
#pragma omp critical
      buffent = current++;
      rd.buffer_sphere(buff, buffent, _pts+3*i, colour(i,defcol,data,cs,dataopac), 0 );
    }
}


/** read in the point file */
bool PPoint :: read( const char *fname )
{
  const int bufsize=1024;
  char      buff[bufsize];

#if USE_OLD
  gzFile in = openFile( fname, "pts" );
  if ( gzgets(in, buff, bufsize) == Z_NULL ) throw 1;
#else
  AsciiReader asciirdr;
  string fn(fname);
  auto pos = fn.rfind(".pts");
  if( pos==string::npos || pos != fn.size()-4 ) {
    if( fn.back() != '.' ) fn += ".";
    fn += "pts";
  }
  asciirdr.read_file(fn);
  asciirdr.get_line( 0, buff, bufsize );
#endif

  if ( sscanf( buff, "%d", &_n ) != 1 ) throw 2;

#if USE_OLD
  gzFileBuffer file(in);
#endif
  
  if ( _base1 ) _n++;					// add initial bogus point
  _pts = (GLfloat *)malloc(_n*3*sizeof(GLfloat));

  GLfloat min[3], max[3];

#pragma omp parallel for private(buff) 
  for ( int i=0; i<3*_n; i+=3 ) {

    if ( i==3 && _base1 ) {				// copy the first point
      for ( int j=0; j<3; j++ )
        _pts[i+j] = _pts[i+j-3];
      continue;
    }
#if USE_OLD
    if ( file.gets(buff, bufsize) == Z_NULL ) throw 2;
#else
    asciirdr.get_line( i/3+1, buff, bufsize );
#endif

    if ( sscanf( buff, "%f %f %f", _pts+i, _pts+i+1, _pts+i+2 ) < 3 )
      throw 3;
  }

  for( int i=0; i<3*_n; i+=3 ) {
    for ( int ti=0; ti<3; ti++ ) {
      if ( !i || _pts[i+ti]>max[ti] ) max[ti] = _pts[i+ti];
      if ( !i || _pts[i+ti]<min[ti] ) min[ti] = _pts[i+ti];
    }
  }
  
  // centre the model about the origin and find maximum distance from origin
  for ( int ti=0; ti<3; ti++ ) _offset[ti] = (min[ti]+max[ti])/2;

#if USE_OLD
  gzclose(in);
#endif

  _allvis.resize( _n );
  _allvis.assign( _n, true );

  return true;
}

#ifdef USE_HDF5
bool PPoint :: read(hid_t hdf_file)
{
  ch5_dataset info;
  if (ch5m_pnts_get_info(hdf_file, &info)) return false;
  
  _pts = (GLfloat*) malloc(info.count * info.width * sizeof(GLfloat));
  _n   = info.count;
  if (ch5m_pnts_get_all(hdf_file, _pts)) {
    free(_pts);
    return false;
  }
  
  // Centre model on origin
  for (int i = 0; i < _n * 3; i += 3) {
    for (int j = 0; j < 3; j++) {
      if (i == 0 || _pts[i+j] > _max[j]) _max[j] = _pts[i+j];
      if (i == 0 || _pts[i+j] < _min[j]) _min[j] = _pts[i+j];
    }
  }
  for (int i = 0; i < 3; i++) _offset[i] = (_min[i]+_max[i])/2;
  
  _allvis.resize(info.count);
  _allvis.assign(info.count, true );
  
  return true;
}
#endif


/** add some points
 *
 *  \param n number to add
 *  \param p points stored in array as x1 y1 z1 x2 y2 z2 ... xn yn zn
 */
void
PPoint :: add( const GLfloat *p, int n )
{
  _n += n;
  _pts = (GLfloat *)realloc( _pts, _n*3*sizeof(GLfloat) );
  memcpy( _pts+3*(_n-n), p, n*3*sizeof(GLfloat) );
  
  // Centre model on origin
  for (int i = (_n-n)*3; i < _n * 3; i += 3) {
    for (int j = 0; j < 3; j++) {
      if (i == 0 || _pts[i+j] > _max[j]) _max[j] = _pts[i+j];
      if (i == 0 || _pts[i+j] < _min[j]) _min[j] = _pts[i+j];
    }
  }
  for (int i = 0; i < 3; i++) _offset[i] = (_min[i]+_max[i])/2;

  _allvis.resize( _n );
  _allvis.assign( _n, true );
  _visible = &_allvis;
  setVis( true );
}


/** make the points dynamic 
 *
 *  \param fn  the name of the dynamic points file
 *  \param ntm the number of times currently loaded
 *
 *  \retval 0 success
 */
int
PPoint :: dynamic( const char *fn, int ntm )
{
  DataOnDemand<float>* newDynPt = new DataOnDemand<float>( fn, _n );

  if( ntm==1 || ntm==0 || ( newDynPt->max_tm()+1==ntm ) ) {
    delete _dynPt;
    _dynPt = newDynPt; 
    return 0;
  } else 
    throw FrameMismatch( ntm, newDynPt->max_tm()+1 );
}


/** set points to the current time 
 *
 *  \param a the time
 *
 *  \note if the requested time is greater than the maximum and the 
 *        maximum is greater than 0, it is an error
 */
void PPoint :: time( int a )
{
  if( _dynPt==NULL || a==_tm || a>_dynPt->max_tm() ) return;

  if( (_pts = _dynPt->slice( a )) ) 
    _tm = a;
}
