#include "DrawingObjects.h"

const int quad_iso_table[][11] = {
    { 0 },
    { 1, 2, 0, 1, 0, 3 },
    { 1, 2, 0, 1, 1, 2 },
    { 1, 2, 0, 3, 1, 2 },
    { 1, 2, 1, 2, 2, 3 },
    { 2, 2, 0, 1, 2, 3, 2, 1, 2, 0, 3 },
    { 1, 2, 0, 1, 2, 3 },
    { 1, 2, 0, 3, 2, 3 },
    { 1, 2, 0, 3, 2, 3 },
    { 1, 2, 0, 1, 2, 3 },
    { 2, 2, 0, 1, 2, 3, 2, 1, 2, 0, 3 },
    { 1, 2, 1, 2, 2, 3 },
    { 1, 2, 0, 3, 1, 2 },
    { 1, 2, 0, 1, 1, 2 },
    { 1, 2, 0, 1, 0, 3 },
    { 0 }
};


/** draw one Quadrilaterals and do not do glBegin/glEnd
 *
 *  \param p0       index of first point to draw
 *  \param colour   colour to use if no data
 *  \param cs       colour scale
 *  \param data     data associated with nodes, element if flat (NULL for no data display)
 *  \param dataopac data opacity
 *  \param nrml     if flat, element normal, else vertex normals  (NULL for none)
 *  \param buffer   buffer entry in which to place data
 *  \param surf     surface we are in
 *  \param flat     flat shading
 */
int Quadrilateral::buffer( int p0, GLfloat *c, Colourscale* cs,
                          DATA_TYPE* data, dataOpac* dataopac,
                          const GLfloat* nrml,  GLfloat *buffer, 
                                                  int surf, bool flat )
{
  if ( p0>=_n ) return 0;

  int i = 4*p0;
  if ( !_pt->vis(_node[i])   || !_pt->vis(_node[i+1]) || 
       !_pt->vis(_node[i+2]) || !_pt->vis(_node[i+3])    )
    return 0;

  GLfloat n[3] = { 0, 0, 0 };
  for ( int j=0; j<3; j++ ) {

    int node = _node[i+j];
    const GLfloat *col = colour( flat?p0:node, c, data, cs, dataopac );  
    buff_vtx( buffer+j*VBO_ELEM_SZ, _pt->pt(node), nrml?(flat?nrml:nrml+node*3):n, col, surf );

    node = _node[i+(j+2)%4];
    col =  colour( flat?p0:node, c, data, cs, dataopac );  
    buff_vtx( buffer+(j+3)*VBO_ELEM_SZ, _pt->pt(node), nrml?(flat?nrml:nrml+node*3):n, col, surf );
  }
  return 2;
}


/** read in the triangle file */
bool Quadrilateral :: read( const char *fname )
{
  gzFile in;

  try {
    in = openFile( fname, "tris" );
  } catch (...) { return false; }

  const int bufsize=1024;
  char      buff[bufsize];
  int       nold=_n, nele;

  while ( gzgets(in, buff, bufsize) != Z_NULL ) {
    sscanf( buff, "%d", &nele );
    _n += nele;
    _node  = (int *)realloc( _node, _n*3*sizeof(int) );
    _nrml = (GLfloat*)realloc( _nrml, _n*sizeof(GLfloat)*3 );
    for ( int i=_n-nele; i<_n; i++ ) {
      gzgets(in, buff, bufsize);
      sscanf( buff, "%d %d %d %*d", _node+3*i, _node+3*i+1, _node+3*i+2 );
    }
  }
  gzclose(in);

  try {
    read_normals( nold, _n-1, fname );
  } catch (...) { compute_normals( nold, _n-1 ); }

  return true;
}


/** increase the number of triangles
 *
 * \param fn   file name with elements to be added
 *
 * \return true if successful
 */
bool Quadrilateral :: add( const char* fn )
{
 return false;
}



/** for a surface with ordered node numbers, compute the normals
 *
 * \param e0 first element
 * \param e1 last element
 */
void Quadrilateral::compute_normals(int e0, int e1)
{
  if (_nrml==NULL) _nrml=new GLfloat[3*e1+3];
  for (int e=e0; e<=e1; e++ ) {

    const int* ele = _node+3*e;
    Vector3D<GLfloat> a(_pt->pt(ele[0]));
    Vector3D<GLfloat> b(_pt->pt(ele[1]));
    Vector3D<GLfloat> c(_pt->pt(ele[2]));

    b -= a;
    b = b.Cross(c-a);
    b.Normalize();
    _nrml[3*e]   = b.X();
    _nrml[3*e+1] = b.Y();
    _nrml[3*e+2] = b.Z();
  }
}



const int* 
Quadrilateral::iso_polys(unsigned int index)
{
  return quad_iso_table[index];
}
